@@ -19,7 +19,6 @@
 import argparse
 import os
 import sys
-import socket
 import subprocess
 import time
 import logging
@@ -34,6 +33,7 @@ from scoop.launch.brokerLaunch import localBroker, remoteBroker
 from .broker.structs import BrokerInfo
 import scoop
 
+# TODO: Is this useful anymore?
 try:
     signal.signal(signal.SIGQUIT, utils.KeyboardInterruptHandler)
 except AttributeError:
@@ -47,15 +47,17 @@ class ScoopApp(object):
 
     def __init__(self, hosts, n, b, verbose, python_executable,
             externalHostname, executable, arguments, tunnel, path, debug,
-            nice, env, profile, pythonPath, prolog, backend):
+            nice, env, profile, pythonPath, prolog, backend, rsh,
+            ssh_executable):
         # Assure setup sanity
-        assert type(hosts) == list and hosts, ("You should at least "
-                                               "specify one host.")
-        self.workersLeft = n
+        assert type(hosts) == list and hosts, (
+            "You should at least specify one host.")
+        self.workersLeft = n if n > 0 else len(hosts)
         self.createdSubprocesses = []
 
         # launch information
         self.python_executable = python_executable[0]
+        self.ssh_executable = ssh_executable
         self.pythonpath = pythonPath
         self.prolog = prolog
         self.n = n
@@ -69,11 +71,12 @@ class ScoopApp(object):
         self.nice = nice
         self.profile = profile
         self.backend = backend
+        self.rsh = rsh
         self.errors = None
 
         # Logging configuration
-        if self.verbose > 2:
-            self.verbose = 2
+        if self.verbose > 3:
+            self.verbose = 3
 
         scoop.logger = utils.initLogging(
             verbosity=self.verbose,
@@ -113,16 +116,18 @@ class ScoopApp(object):
         # Logging of worker distribution warnings
         maximumWorkers = sum(host[1] for host in hosts)
         if self.n > maximumWorkers:
-            scoop.logger.debug("The -n flag is set at {0} workers, which is higher "
-                           "than the maximum number of workers ({1}) specified "
-                           "by the hostfile.\nThis behavior may degrade the "
-                           "performances of scoop for cpu-bound operations."
-                           "".format(qty, maximumWorkers))
-        elif self.n < maximumWorkers:
-            scoop.logger.debug("The -n flag is set at {0} workers, which is lower "
-                           "than the maximum number of workers ({1}) specified "
-                           "by the hostfile."
-                           "".format(qty, maximumWorkers))
+            scoop.logger.debug(
+                "The -n flag is set at {0} workers, which is higher than the "
+                "maximum number of workers ({1}) specified by the hostfile.\n"
+                "This behavior may degrade the performances of scoop for "
+                "cpu-bound operations.", self.n, maximumWorkers
+            )
+        elif self.n < maximumWorkers and self.n > 0:
+            scoop.logger.debug(
+                "The -n flag is set at {0} workers, which is lower than the "
+                "maximum number of workers ({1}) specified by the hostfile.",
+                self.n, maximumWorkers
+            )
 
         # Display
         self.showHostDivision(headless=not executable)
@@ -157,7 +162,7 @@ class ScoopApp(object):
 
         # If specified amount of workers if lower than sum of each specified.
         elif qty < maximumWorkers:
-            
+
             while qty < maximumWorkers:
                 maximumWorkers -= hosts[-1][1]
                 if qty > maximumWorkers:
@@ -176,16 +181,14 @@ class ScoopApp(object):
                             "broker: Unresolvable hostname or IP address.\n "
                             "Please specify your externally routable hostname "
                             "or IP using the --external-hostname parameter or "
-                            " use the --tunnel flag.")
-
-        hosts.reverse()
+                            "use the --tunnel flag.")
         return hosts
 
     def showHostDivision(self, headless):
-        """Show the worker distribution over the hosts"""
+        """Show the worker distribution over the hosts."""
         scoop.logger.info('Worker distribution: ')
-        for worker, number in reversed(self.worker_hosts):
-            first_worker = (worker == self.worker_hosts[-1][0])
+        for worker, number in self.worker_hosts:
+            first_worker = (worker == self.worker_hosts[0][0])
             scoop.logger.info('   {0}:\t{1} {2}'.format(
                 worker,
                 number - 1 if first_worker or headless else str(number),
@@ -193,15 +196,13 @@ class ScoopApp(object):
                 )
             )
 
-    def _addWorker_args(self, workerinfo):
+    def _setWorker_args(self, origin):
         """Create the arguments to pass to the addWorker call.
             The returned args and kwargs must ordered/named according to the namedtuple
-            in LAUNCH_HOST_CLASS.LAUNCHING_ARGUMENTS
+            in LAUNCH_HOST_CLASS.LAUNCHING_ARGUMENTS .
 
-            both args and kwargs are supported for full flexibilty,
-            but usage of kwargs only is strongly advised
-
-            workerinfo is a dict with information that can be used to start the worker
+            both args and kwargs are supported for full flexibility,
+            but usage of kwargs only is strongly advised.
         """
         args = []
         kwargs = {
@@ -211,7 +212,7 @@ class ScoopApp(object):
             'nice': self.nice,
             'pythonExecutable': self.python_executable,
             'size': self.n,
-            'origin': self.workersLeft == 1,
+            'origin': origin,
             'brokerHostname': self.externalHostname,
             'brokerPorts': (self.brokers[0].brokerPort,
                             self.brokers[0].infoPort),
@@ -224,27 +225,26 @@ class ScoopApp(object):
         }
         return args, kwargs
 
-    def addWorkerToHost(self, workerinfo):
-        """Adds a worker to current host"""
-        hostname = workerinfo['hostname']
+    def setWorkerInfo(self, hostname, workerAmount, origin):
+        """Sets the worker information for the current host."""
 
         scoop.logger.debug('Initialising {0}{1} worker {2} [{3}].'.format(
             "local" if hostname in utils.localHostnames else "remote",
-            " origin" if self.workersLeft == 1 else "",
+            " origin" if origin else "",
             self.workersLeft,
             hostname,
             )
         )
 
-        add_args, add_kwargs = self._addWorker_args(workerinfo)
-        self.workers[-1].addWorker(*add_args, **add_kwargs)
+        add_args, add_kwargs = self._setWorker_args(origin)
+        self.workers[-1].setWorker(*add_args, **add_kwargs)
+        self.workers[-1].setWorkerAmount(workerAmount)
 
     def run(self):
         """Launch the broker(s) and worker(s) assigned on every hosts."""
-        # Launch the brokers
+        # Launch the broker(s)
         for hostname, nb_brokers in self.broker_hosts:
             for ind in range(nb_brokers):
-                # Launching the broker(s)
                 if self.externalHostname in utils.localHostnames:
                     self.brokers.append(localBroker(
                         debug=self.debug,
@@ -258,6 +258,8 @@ class ScoopApp(object):
                         debug=self.debug,
                         nice=self.nice,
                         backend=self.backend,
+                        rsh=self.rsh,
+                        ssh_executable=self.ssh_executable,
                     ))
 
         # Share connection information between brokers
@@ -276,17 +278,16 @@ class ScoopApp(object):
                 broker.sendConnect(connect_data)
 
         # Launch the workers
+        shells = []
+        origin_launched = False
         for hostname, nb_workers in self.worker_hosts:
-            self.workers.append(self.LAUNCH_HOST_CLASS(hostname))
+            self.workers.append(self.LAUNCH_HOST_CLASS(hostname, self.rsh,
+                                                       self.ssh_executable))
             total_workers_host = min(nb_workers, self.workersLeft)
-            for worker_idx_host in range(total_workers_host):
-                workerinfo = {
-                    'hostname': hostname,
-                    'total_workers_host': total_workers_host,
-                    'worker_idx_host': worker_idx_host,
-                }
-                self.addWorkerToHost(workerinfo)
-                self.workersLeft -= 1
+
+            self.setWorkerInfo(hostname, total_workers_host, not origin_launched)
+            origin_launched = True
+            self.workersLeft -= total_workers_host
 
             # Launch every workers at the same time
             scoop.logger.debug(
@@ -295,31 +296,24 @@ class ScoopApp(object):
                     self.workers[-1].getCommand(),
                 )
             )
-            shells = self.workers[-1].launch(
+            shells.append(self.workers[-1].launch(
                 (self.brokers[0].brokerPort,
                  self.brokers[0].infoPort)
                     if self.tunnel else None,
-                stdPipe=not self.workers[-1].isLocal(),
-            )
+            ))
             if self.workersLeft <= 0:
                 # We've launched every worker we needed, so let's exit the loop
-                rootProcess = shells[-1]
                 break
+        rootProcess = shells[0][0]
 
         # Wait for the root program
-        if self.workers[-1].isLocal():
-            self.errors = self.workers[-1].subprocesses[-1].wait()
-        else:
-            # Process stdout first, then the whole stderr at the end
-            for outStream, inStream in [(sys.stdout, rootProcess.stdout),
-                                        (sys.stderr, rootProcess.stderr)]:
-                data = inStream.read(1)
-                while len(data) > 0:
-                    # Should not rely on utf-8 codec
-                    outStream.write(data.decode("utf-8"))
-                    outStream.flush()
-                    data = inStream.read(1)
-            self.errors = rootProcess.wait()
+        try:
+            if self.workers[0].isLocal():
+                self.errors = self.workers[0].subprocesses[0].wait()
+            else:
+                self.errors = rootProcess.wait()
+        except KeyboardInterrupt:
+            pass
         scoop.logger.info('Root process is done.')
         return self.errors
 
@@ -327,7 +321,7 @@ class ScoopApp(object):
         """Subprocess cleanup."""
         # Give time to flush data if debug was on
         if self.debug:
-            time.sleep(5)
+            time.sleep(10)
 
         # Terminate workers
         for host in self.workers:
@@ -393,6 +387,14 @@ def makeParser():
                         type=int,
                         default=1,
                         metavar="NumberOfBrokers")
+    parser.add_argument('--rsh',
+                        help="Use RSH instead of SSH for the launching process. "
+                             "Not compatible with --tunnel flag.",
+                        action='store_true')
+    parser.add_argument('--ssh-executable',
+                        help="Name of the ssh executable. (default: 'ssh')",
+                        default="ssh",
+                        metavar="SSHExecutable")
     parser.add_argument('--tunnel',
                         help="Activate ssh tunnels to route toward the broker "
                              "sockets over remote connections (may eliminate "
@@ -454,12 +456,16 @@ def main():
 
     # Get a list of resources to launch worker(s) on
     hosts = utils.getHosts(args.hostfile, args.hosts)
+
     if args.n:
         n = args.n
     else:
         n = utils.getWorkerQte(hosts)
-    assert n > 0, ("Scoop couldn't determine the number of worker to start.\n"
-                   "Use the '-n' flag to set it manually.")
+    assert n >= 0, (
+            "Scoop couldn't determine the number of worker to start.\n"
+            "Use the '-n' flag to set it manually."
+    )
+
     if not args.external_hostname:
         args.external_hostname = [utils.externalHostname(hosts)]
 
@@ -471,7 +477,8 @@ def main():
                             args.executable, args.args, args.tunnel,
                             args.path, args.debug, args.nice,
                             utils.getEnv(), args.profile, args.pythonpath[0],
-                            args.prolog[0], args.backend)
+                            args.prolog[0], args.backend, args.rsh,
+                            args.ssh_executable)
 
     rootTaskExitCode = False
     interruptPreventer = Thread(target=thisScoopApp.close)
