@@ -82,6 +82,10 @@ def funcLambda(n):
     return sum(result)
 
 
+def funcWithKW(n, **kwargs):
+    return kwargs
+
+
 def funcLambdaSubfuncNotGlobal(n):
     """Tests a lambda function containing a call to a function that is not in
     the globals()."""
@@ -165,7 +169,6 @@ def funcMapScan(l):
     resultat = futures.mapScan(func4,
                                operator.add,
                                l)
-    time.sleep(0.3)
     _control.execQueue.socket.pumpInfoSocket()
     return resultat
 
@@ -174,7 +177,6 @@ def funcMapReduce(l):
     resultat = futures.mapReduce(func4,
                                  operator.add,
                                  l)
-    time.sleep(0.3)
     _control.execQueue.socket.pumpInfoSocket()
     return resultat
 
@@ -185,7 +187,6 @@ def funcDoubleMapReduce(l):
     resultat2 = futures.mapReduce(func4,
                                  operator.add,
                                  l)
-    time.sleep(0.3)
     _control.execQueue.socket.pumpInfoSocket()
     return resultat == resultat2
 
@@ -243,6 +244,13 @@ def funcIter(n):
     return sum(result)
 
 
+def funcKeywords(n, **kwargs):
+    task = futures.submit(funcWithKW, n, **kwargs)
+    futures.wait([task], return_when=futures.ALL_COMPLETED)
+    result = task.result()
+    return result
+
+
 def main(n):
     task = futures.submit(func0, n)
     futures.wait([task], return_when=futures.ALL_COMPLETED)
@@ -250,7 +258,7 @@ def main(n):
     return result
 
 
-def main_simple(n):
+def mainSimple(n):
     task = futures.submit(func3, n)
     futures.wait([task], return_when=futures.ALL_COMPLETED)
     result = task.result()
@@ -261,20 +269,20 @@ def submit_get_queues_size(n):
     task = futures.submit(func4, n)
     result = task.result()
     return [
-        len(scoop._control.execQueue.inprogress),
-        len(scoop._control.execQueue.ready),
-        len(scoop._control.execQueue.movable),
-        len(scoop._control.futureDict) - 1, # - 1 because the current function is a future too
+        len(_control.execQueue.inprogress),
+        len(_control.execQueue.ready),
+        len(_control.execQueue.movable),
+        len(_control.futureDict) - 1, # - 1 because the current function is a future too
     ]
 
 
 def map_get_queues_size(n):
     result = list(map(func4, [n for n in range(n)]))
     return [
-        len(scoop._control.execQueue.inprogress),
-        len(scoop._control.execQueue.ready),
-        len(scoop._control.execQueue.movable),
-        len(scoop._control.futureDict) - 1, # - 1 because the current function is a future too
+        len(_control.execQueue.inprogress),
+        len(_control.execQueue.ready),
+        len(_control.execQueue.movable),
+        len(_control.futureDict) - 1, # - 1 because the current function is a future too
     ]
 
 
@@ -297,7 +305,7 @@ class TestScoopCommon(unittest.TestCase):
 
     def multiworker_set(self):
         global subprocesses
-        worker = subprocess.Popen([sys.executable, "-m", "scoop.bootstrap",
+        worker = subprocess.Popen([sys.executable, "-m", "scoop.bootstrap.__main__",
         "--brokerHostname", "127.0.0.1", "--taskPort", "5555",
         "--metaPort", "5556", "tests.py"])
         subprocesses.append(worker)
@@ -305,17 +313,19 @@ class TestScoopCommon(unittest.TestCase):
 
     def setUp(self):
         global subprocesses
-        # Start the server
-        self.server = subprocess.Popen([sys.executable, "-m", "scoop.broker",
-        "--tPort", "5555", "--mPort", "5556"])
         import socket, datetime, time
+
+        # Start the server
+        self.server = subprocess.Popen([sys.executable, "-m", "scoop.broker.__main__",
+        "--tPort", "5555", "--mPort", "5556"])
         s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         begin = datetime.datetime.now()
         while not port_ready(5555, s):
             if (datetime.datetime.now() - begin > datetime.timedelta(seconds=3)):
                 raise Exception('Could not start server!')
-            pass
         subprocesses.append(self.server)
+
+        # Setup worker environment
         scoop.IS_RUNNING = True
         scoop.IS_ORIGIN = True
         scoop.WORKER_NAME = 'origin'.encode()
@@ -329,20 +339,36 @@ class TestScoopCommon(unittest.TestCase):
         scoop.VALID = True
         scoop.DEBUG = False
         scoop.SIZE = 2
-        scoop._control.execQueue = FutureQueue()
+        _control.execQueue = FutureQueue()
 
     def tearDown(self):
         global subprocesses
-        scoop._control.futureDict.clear()
-        try: self.w.kill()
-        except: pass
+        import socket, datetime, time
+        _control.execQueue.shutdown()
+        del _control.execQueue
+        _control.futureDict.clear()
+        try:
+            self.w.terminate()
+            self.w.wait()
+        except:
+            pass
         # Destroy the server
         if self.server.poll() == None:
-            try: self.server.kill()
-            except: pass
+            try:
+                self.server.terminate()
+                self.server.wait()
+            except:
+                pass
         # Stabilise zmq after a deleted socket
         del subprocesses[:]
-        time.sleep(0.1)
+
+        # Wait for the previous server to be correctly terminated
+        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        begin = datetime.datetime.now()
+        while port_ready(5555, s):
+            if (datetime.datetime.now() - begin > datetime.timedelta(seconds=3)):
+                raise Exception('Could not terminate server!')
+        s.close()
 
 
 class TestMultiFunction(TestScoopCommon):
@@ -415,17 +441,17 @@ class TestMultiFunction(TestScoopCommon):
     def test_execQueue_multiworker(self):
         self.w = self.multiworker_set()
         result = futures._startup(func0, 6)
-        self.assertEqual(len(scoop._control.execQueue.inprogress), 0)
-        self.assertEqual(len(scoop._control.execQueue.ready), 0)
-        self.assertEqual(len(scoop._control.execQueue.movable), 0)
-        self.assertEqual(len(scoop._control.futureDict), 0)
+        self.assertEqual(len(_control.execQueue.inprogress), 0)
+        self.assertEqual(len(_control.execQueue.ready), 0)
+        self.assertEqual(len(_control.execQueue.movable), 0)
+        self.assertEqual(len(_control.futureDict), 0)
 
     def test_execQueue_uniworker(self):
         result = futures._startup(func0, 6)
-        self.assertEqual(len(scoop._control.execQueue.inprogress), 0)
-        self.assertEqual(len(scoop._control.execQueue.ready), 0)
-        self.assertEqual(len(scoop._control.execQueue.movable), 0)
-        self.assertEqual(len(scoop._control.futureDict), 0)
+        self.assertEqual(len(_control.execQueue.inprogress), 0)
+        self.assertEqual(len(_control.execQueue.ready), 0)
+        self.assertEqual(len(_control.execQueue.movable), 0)
+        self.assertEqual(len(_control.futureDict), 0)
 
     def test_execQueue_submit_uniworker(self):
         result = futures._startup(submit_get_queues_size, 6)
@@ -461,7 +487,6 @@ class TestMultiFunction(TestScoopCommon):
             "Buffers are not empty after future completion"
         )
 
-
     def test_partial(self):
         """This function removes some attributes (such as __name__)."""
         from functools import partial
@@ -473,7 +498,7 @@ class TestSingleFunction(TestMultiFunction):
     def __init__(self, *args, **kwargs):
         # Parent initialization
         super(TestSingleFunction, self).__init__(*args, **kwargs)
-        self.main_func = main_simple
+        self.main_func = mainSimple
         self.small_result = 30
         self.large_result = 2870
 
@@ -505,6 +530,10 @@ class TestApi(TestScoopCommon):
         result = futures._startup(funcLambda, 30)
         self.assertEqual(result, 9455)
 
+    def test_submit_with_keyword(self):
+        result = futures._startup(funcKeywords, 2, kwarg=3.1415926)
+        self.assertEqual(result, { "kwarg": 3.1415926} )
+
     # This test is complex to handle and has many implications
     # Bundle a closure with the future?
     # How to manage side-effects of variables in closure?
@@ -642,4 +671,4 @@ if __name__ == '__main__' and os.environ.get('IS_ORIGIN', "1") == "1":
     else:
         unittest.main()
 elif __name__ == '__main__':
-    futures._startup(main_simple)
+    futures._startup(mainSimple)
